/*
 * This file is part of jdbw (http://code.google.com/p/jdbw/).
 * 
 * jdbw is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 * 
 * Copyright (C) 2007-2012 Martin Berglund
 */
package com.googlecode.jdbw.jorm;

import java.lang.reflect.InvocationHandler;
import java.lang.reflect.Method;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.UUID;

class EntityProxy<U, T extends JORMEntity<U>> implements InvocationHandler {

    private static final Map<Class, Map<String, Integer>> INDEX_REFERENCE = new HashMap<Class, Map<String, Integer>>();
    
    private static <U, T extends JORMEntity<U>> void indexClass(Class<T> entityClass, ClassTableMapping classTableMapping) {
        synchronized(INDEX_REFERENCE) {
            if(INDEX_REFERENCE.containsKey(entityClass))
                return;
            
            Map<String, Integer> classFieldMap = new HashMap<String, Integer>();
            int counter = 0;
            for(String fieldName: classTableMapping.getFieldNames(entityClass)) {
                classFieldMap.put(fieldName, counter++);
            }
            INDEX_REFERENCE.put(entityClass, classFieldMap);
        }
    }
    
    private static <U, T extends JORMEntity<U>> int getNumberOfFields(Class<T> entityClass) {
        synchronized(INDEX_REFERENCE) {
            return INDEX_REFERENCE.get(entityClass).size();
        }
    }
    
    private static <U, T extends JORMEntity<U>> int getFieldIndex(Class<T> entityClass, String field) {
        synchronized(INDEX_REFERENCE) {
            return INDEX_REFERENCE.get(entityClass).containsKey(field) ? 
                        INDEX_REFERENCE.get(entityClass).get(field) : -1;
        }
    }
    
    private final Class<T> entityClass;
    private final ClassTableMapping mapping;    
    private final U id;
    private final Object[] values;
    
    EntityProxy(Class<T> entityClass, ClassTableMapping mapping, U id, Object[] initData) {
        indexClass(entityClass, mapping);
        this.entityClass = entityClass;
        this.mapping = mapping;
        this.id = id;
        this.values = Arrays.copyOf(initData, getNumberOfFields(entityClass));
    }

    @Override
    public synchronized Object invoke(Object proxy, Method method, Object[] args) throws Throwable {
        if(method.getName().equals("getId") && args == null) {
            return id;
        }
        else if(method.getName().equals("__underlying_proxy") && args == null) {
            return this;
        }
        else if(method.getName().equals("toString") && args == null) {
            return toString();
        }
        else if(method.getName().equals("hashCode") && args == null) {
            return hashCode();
        }
        else if(method.getName().equals("equals") && 
                method.getParameterTypes().length == 1 && 
                method.getParameterTypes()[0] == Object.class) {
            
            return equals(args[0]);
        }
        else if(method.getName().startsWith("get") && method.getName().length() > 3 && method.getParameterTypes().length == 0) {
            String asFieldName = Character.toLowerCase(method.getName().charAt(3)) + method.getName().substring(4);
            if(getFieldIndex(entityClass, asFieldName) == -1) {
                throw new IllegalStateException("In method call to " + entityClass.getSimpleName() + 
                        "." + method.getName() + ", couldn't find field " + asFieldName + " in " +
                        "EntityProxy");
            }
            return convertToReturnType(method.getReturnType(), getValue(asFieldName));
        }
        else if(method.getName().startsWith("set") && method.getName().length() > 3 && method.getParameterTypes().length == 1) {
            String asFieldName = Character.toLowerCase(method.getName().charAt(3)) +
                                    method.getName().substring(4);
            if(getFieldIndex(entityClass, asFieldName) == -1) {
                throw new IllegalStateException("In method call to " + entityClass.getSimpleName() + 
                        "." + method.getName() + ", couldn't find field " + asFieldName + " in " +
                        "EntityProxy");
            }
            setValue(asFieldName, args[0]);
            if(method.getReturnType() != Void.class)
                return proxy;
            else
                return null;
        }
        else {
            throw new IllegalArgumentException("EntityProxy doesn't know how to handle a call to " +
                    entityClass.getSimpleName() + "." + method.getName());
        }
    }

    synchronized void populate(Object[] row) {
        System.arraycopy(row, 1, values, 0, values.length); //Remember, the id is at index 0, we don't want to copy that!
    }
    
    synchronized Object getValue(String columnName) {
        return values[getFieldIndex(entityClass, columnName)];
    }
    
    private <T> T convertToReturnType(Class<T> returnType, Object object) {
        if(object == null)
            return null;
        
        if(returnType.isAssignableFrom(object.getClass()))
            return (T)object;
        
        if(returnType.equals(UUID.class) && object instanceof String)
            return (T)UUID.fromString((String)object);
        
        if(returnType.isPrimitive()) {            
            return (T)object; //Let java autoboxing do its thing
        }
        throw new IllegalArgumentException("Cannot typecast " + object.getClass().getName() + " to " + returnType.getName());
    }

    private void setValue(String columnName, Object value) {
        values[getFieldIndex(entityClass, columnName)] = value;
    }
    
    Class<T> getEntityType() {
        return entityClass;
    }
    
    static interface Resolver<U, T extends JORMEntity<U>> {
        EntityProxy<U, T> __underlying_proxy();
    }

    @Override
    public synchronized String toString() {
        return entityClass.getSimpleName() + "(id=" + id + ")" + values;
    }

    @Override
    public boolean equals(Object obj) {
        if(obj == null) {
            return false;
        }
        if(this == obj) {
            return true;
        }
        if(!entityClass.isAssignableFrom(obj.getClass()) &&
                obj.getClass() != getClass()) {
            return false;
        }
        
        U otherId;
        if(entityClass.isAssignableFrom(obj.getClass())) {
            otherId = ((JORMEntity<U>)obj).getId();
        }
        else {
            final EntityProxy<U, T> other = (EntityProxy<U, T>) obj;
            if(this.entityClass != other.entityClass && (this.entityClass == null || !this.entityClass.equals(other.entityClass))) {
                return false;
            }
            otherId = other.id;
        }        
        if(this.id != otherId && (this.id == null || !this.id.equals(otherId))) {
            return false;
        }
        return true;
    }

    @Override
    public int hashCode() {
        return entityClass.getSimpleName().hashCode() + id.hashCode();
    }
}
